#include "shmem.h"
#include "filed.h"
#include "main.h"
#include "log.h"
#include "externs.h"
#include "sys.h"
#include "dbq.h"
#include "parse.h"
#include "url.h"
#include "socket.h"
#include "session.h"
#include "stationd.h"
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>

int grow_shmem(void) {
	/* this attempts to make the shared memory segment 16kb bigger */
	/* than we need. returns 1 if success, 0 if fail */
	int need;

	need = get_avail_shmem() - sizeof(struct msg);
	if(need < 0) {
		need = need * -1;
		need += get_shmmax();
		need += 256 * 1024;
		return(put_shmmax(need));
	}
	
	return(1);
}
	
void display_shmem(void) {
	int need;

	printf("This version of filed was compiled for %d files\n",SHMEM_MAX_FILES);
	printf("Size of shared memory segment:\t%d\n",sizeof(struct msg));
	printf("Size of shared memory total:\t%d\n",get_shmmax());
	printf("Size of shared memory avail:\t%d\n",get_avail_shmem());
	need = get_avail_shmem() - sizeof(struct msg);
	if(need < 0) {
		need = need * -1;
		printf("Size of shared memory needed:\t%d\n",need);
	}
}


void bind_shmem(void) {
	struct msg *msg;
	key_t shmkey;
	int i;
	msg = NULL;
	
	shmkey = (key_t) get_one_data_int(filed->cf,"shmem-key");
	dlog(DEBUG_IPC,"shared memory segment: %d",shmkey);
	
	if((filed->shmid = shmget(shmkey,sizeof(struct msg),PERMS | IPC_CREAT)) < 0) {
		filed->shmid = 0;
		elog("bind_shmem(): unable to get shared memory segment(%d)",shmkey);
		die();
	}

	if((msg = (struct msg *) shmat(filed->shmid,(char *) 0, 0)) == (struct msg *) -1) {
		filed->shmid = 0;
		elog("Unable to attach shared memory segment");
		die();
	}

	
	if(!(msg->master_flags & SHMEM_FLAG_READY)) 

	msg->master_flags = SHMEM_FLAG_READY;
	
	for(i=0;i<SHMEM_MAX_FILES;i++) {
		strcpy(msg->from[i],"");
		strcpy(msg->to[i],"");
		msg->id[i] = 0;
		msg->flags[i] = 0;	
	}

	msg->max_entry = 0;	

	for(i=0;i<SHMEM_NUM_QUEUES;i++) {
		strcpy(msg->message[i],"");
		msg->message_flags[i] = 0;
		msg->connection[i] = 0;
		msg->filter[i] = 0;
	}
	msg->max_entry = 0;

	filed->msg = msg;
}


void bind_shmem_client(void) {
	struct msg *msg;
	key_t shmkey;
	msg = NULL;
	
	shmkey = (key_t) get_one_data_int(filed->cf,"shmem-key");
	
	if((filed->shmid = shmget(shmkey,sizeof(struct msg),PERMS)) < 0) {
		filed->shmid = 0;
		elog("bind_shmem_client(): unable to get shared memory segment(%d)",shmkey);
		die();
	}

	if((msg = (struct msg *) shmat(filed->shmid,(char *) 0, 0)) == (struct msg *) -1) {
		filed->shmid = 0;
		elog("Unable to attach shared memory segment");
		die();
	}

	
	filed->msg = msg;
}

void handle_shmem(void) {
	/* take any shared memory segment that has info pending and send it out */
}

void wait_for_db(void) {
	while(!(filed->msg->master_flags & SHMEM_FLAG_DBREADY)) usleep(1);
	return;
}

void wait_for_main(void) {
	while(!(filed->msg->master_flags & SHMEM_FLAG_READY)) usleep(1);
	return;
}

void shmem_db_ready(void) {
	if(!(filed->msg->master_flags & SHMEM_FLAG_DBREADY)) filed->msg->master_flags += SHMEM_FLAG_DBREADY;
}

int shmem_db_free(void) {
	/* return the number of message queues not marked RX or TX */
	struct msg *msg = filed->msg;
	int count = 0;
	int i;

	for(i=0;i<SHMEM_NUM_QUEUES;i++) {
		if(!(msg->message_flags[i] & SHMEM_CLAIMED)) count++;
	}
	
	return(count);
}

int shmem_db_id_present(int id) {
	/* check to see if ID is already present */
	/* used to keep connection-based queries from arriving out of order */
	
	int i;
	struct msg *msg = filed->msg;

	for(i=0;i<SHMEM_NUM_QUEUES;i++) 
		if((msg->connection[i] == id) && (msg->filter > 0) && (msg->message_flags[i] & SHMEM_CLAIMED)) return(1);
	
	return(0);
}

void shmem_db_ins(struct db *db) {
	/* this function finds a free slot 
	 * and inserts our queued query into it 
	 */

	int i;
	struct msg *msg = filed->msg;
	
	if(db->flags & DBFLAG_PROCESSING) return;

	for(i=0;i<SHMEM_NUM_QUEUES;i++) {
		if(!(msg->message_flags[i] & SHMEM_CLAIMED)) {
			/* it's all ours */
			dlog(DEBUG_IPC,"inserting query %d",db->seq);
			msg->message_flags[i] = SHMEM_CLAIMED;
			msg->message_flags[i] += SHMEM_OKTOGO;
			strcpy(msg->message[i],db->query);
			msg->filter[i] = db->filter;
			msg->connection[i] = db->id;
			db->flags += DBFLAG_PROCESSING;
			db->loc = i;
			return;
		}
	}

	return;
}

struct db *shmem_db_get_next(void) {
	/* This is a _SERVER_ thread application */
	static struct db d;
	struct msg *msg = filed->msg;
	int i;

	for(i=0;i<SHMEM_NUM_QUEUES;i++) {
		if(msg->message_flags[i] & SHMEM_OKTOGO) {
			d.id = msg->connection[i];
			d.filter = msg->filter[i];
			d.loc = i;
			d.seq = 0;
			d.flags = msg->message_flags[i];
			d.query = strdup(msg->message[i]);
			d.output = NULL;
			d.prev = NULL;
			d.next = NULL;
			return(&d);
		}
	}
	return(NULL);
}

void shmem_db_set(struct db *db) {
	/* this is a SERVER thread function */

	struct msg *msg = filed->msg;
	int i = db->loc;

	if(db->output) strcpy(msg->message[i],db->output);
	else strcpy(msg->message[i],"");

	msg->message_flags[i] = db->flags;
	msg->connection[i] = db->id;
	msg->filter[i] = db->filter;
}
	
struct db *shmem_db_get_done(void) {
	/*
	 * this function retreives any pending output from
	 * the db communications channel
	 * and sets the DBFLAG_DONE 
	 * if we've got all the return data we're going to get
	 */

	int i;
	struct msg *msg = filed->msg;
	struct db *db;
	
	for(i=0;i<SHMEM_NUM_QUEUES;i++) {
		if((db = dbq_get_slot(i))) 
		if(msg->message_flags[i] & SHMEM_PROCESSED) {
			if(db->output) free(db->output);
			db->output = strdup(msg->message[i]);
			msg->message_flags[i] -= SHMEM_PROCESSED;
			if(msg->message_flags[i] & SHMEM_MORECOMING)
				msg->message_flags[i] += SHMEM_OKTOGO;
			else
				msg->message_flags[i] = 0;

			return(db);
		}
	}
	return(NULL);	
}

char *shmem_msg_flags_string(int flags) {
	static char ret[1024];

	strcpy(ret,"");
	if(flags & SHMEM_CLAIMED) strcat(ret,"CLAIMED ");
	if(flags & SHMEM_PROCESSED) strcat(ret,"PROCESSED ");
	if(flags & SHMEM_OKTOGO) strcat(ret,"OKTOGO ");
	if(flags & SHMEM_MORECOMING) strcat(ret,"MORECOMING ");
	if(flags & SHMEM_DONE) strcat(ret,"DONE ");
	if(flags & SHMEM_ERR) strcat(ret,"ERR ");
	return(ret);
}

void shmem_connection_dump(struct connection *c) {
	/* dump entire shmem to connection */
	char outbuf[1024];
	int i;
	struct msg *msg = filed->msg;
	
	for(i=0;i<SHMEM_NUM_QUEUES;i++) {
		sprintf(outbuf,"%d: %s %d %d [%s]",i,shmem_msg_flags_string(msg->message_flags[i]),msg->filter[i],msg->connection[i],msg->message[i]);
		c_send(c,outbuf);
	}
}
	
int shmem_ping_db_thread(void) {
	struct msg *msg = filed->msg;
	
	if(!msg) return(0);

	if(msg->master_flags & SHMEM_FLAG_DBPING) {
		usleep(1);
		if(msg->master_flags & SHMEM_FLAG_DBPING) return(1);	
	}
	msg->master_flags += SHMEM_FLAG_DBPING;
	return(0);
}
	

void shmem_shutdown(void) {
	/* unattach */
	struct msg *msg = filed->msg;

	if(msg && (msg->master_flags & SHMEM_FLAG_READY))
		msg->master_flags -= SHMEM_FLAG_READY;

	if(filed->shmid)
		shmdt(filed->msg);
	
}

int shmem_check(void) {
	/* check to see if there is another filed running */
	/* first we attach a shared memory segment */
	struct msg *msg;
	key_t shmkey;
	int shmid;
	
	shmkey = (key_t) get_one_data_int(filed->cf,"shmem-key");
	msg = NULL;

	dlog(DEBUG_IPC,"checking for another filed on IPC segment: %d",shmkey);
	
	if((shmid = shmget(shmkey,sizeof(struct msg),PERMS)) < 0) 
		return(0);

	if((msg = (struct msg *) shmat(shmid,(char *) 0, 0)) == (struct msg *) -1) 
		return(0);	

	
	if(msg->master_flags & SHMEM_FLAG_READY) {
		if(!(msg->master_flags & SHMEM_FLAG_PING)) {
			msg->master_flags += SHMEM_FLAG_PING;
			sleep(1);
			if(msg->master_flags & SHMEM_FLAG_PING) {
				shmdt(msg);
				return(0);
					} else {
				shmdt(msg);
				return(1);
				}
			} else {
				shmdt(msg);
				return(0);
			}
	} else {
		shmdt(msg);
		return(0);	
	}
	
}	

void shmem_respond_to_filed_pings(void) {
	struct msg *msg = filed->msg;
	if(msg->master_flags & SHMEM_FLAG_PING) 
		msg->master_flags -= SHMEM_FLAG_PING;
}

void shmem_session_kill(int skey) {
	int i;
	struct msg *msg = filed->msg;

	for(i=0;i<SHMEM_MAX_FILES;i++) {
		if(msg->id[i] == i) {
			shmem_file_remove(i);
		}
	}
}

void shmem_file_add(char *from, char *to, int session_id) {
	/* add an active file */
	struct msg *msg = filed->msg;
	int i;
	
	/* step 1: look and see if anything in current scope is available */

	if(!from || !to) {
		dlog(DEBUG_IPC,"shmem_file_add() called without from or to");
		return;
	}
	
	for(i=0;i<=msg->max_entry;i++) {
		if(!(msg->flags[i] & SHMEM_ACTIVE)) break;
	}
	
	if(i>=msg->max_entry) {
		i = ++msg->max_entry;
		if(i > SHMEM_MAX_FILES) {
			elog("shmem_file_add(): filed ran out of available file slots (compiled in: %d)",SHMEM_MAX_FILES);
			die();	
		}
		dlog(DEBUG_IPC,"shmem_file_add() grew max_entry: %d",msg->max_entry);
	}

	msg->flags[i] = SHMEM_ACTIVE;
	strcpy(msg->from[i],from);
	strcpy(msg->to[i],to);
	msg->id[i] = session_id;
	dlog(DEBUG_IPC,"File [%s] -> [%s] saved in slot %d",from,to,i);
	return;
}
	
void shmem_file_remove(int slot) {
	struct msg *msg = filed->msg;

	msg->id[slot] = 0;
	msg->flags[slot] = 0;
	strcpy(msg->to[slot],"");
	strcpy(msg->from[slot],"");
	
	if(slot == msg->max_entry) {
		for(;msg->max_entry > 0; msg->max_entry--) 
			if(msg->flags[msg->max_entry] & SHMEM_ACTIVE) break;
			
		dlog(DEBUG_IPC,"shmem_file_remove() shrunk max_entry: %d",msg->max_entry);
	}
}
	
void shmem_slot_show(struct connection *c, struct parse *p) {
	char *arg1 = get_token_s(p,2);
	struct msg *msg = filed->msg;
	int i;
	int inuse;

	

	if(!arg1) {
		inuse = 0;
		for(i=0;i<=msg->max_entry;i++) 
			if(msg->flags[i] & SHMEM_ACTIVE) inuse++;
		c_send(c,"OK %d slots in scan %d slots in use",msg->max_entry,inuse);
		return;
	}


	i = atoi(arg1);
	c_send(c,"OK [%d] F: [%s] T: [%s] fl: %s",i,msg->from[i],msg->to[i],shmem_file_flags_string(msg->flags[i]));
	return;
}

char *shmem_file_flags_string(int flags) {
	static char ret[1024];
	strcpy(ret,"");
	
	if(flags & SHMEM_ACTIVE) strcat(ret,"ACTIVE ");
	if(flags & SHMEM_OPENED) strcat(ret,"OPENED ");
	if(flags & SHMEM_TOUCHED) strcat(ret,"TOUCHED ");
	if(flags & SHMEM_STATIOND) strcat(ret, "STATIOND ");
	if(flags & SHMEM_INS_DEL) strcat(ret,"INS_DEL ");
	return(ret);
}
		
void shmem_handle(void) {
	/* 
	 * handle new events in shmem
	 * most of these are callouts to file.c
 	 *
	 */
	
	int i;
	struct msg *msg = filed->msg;
	
	for(i=0;i<=msg->max_entry;i++) {
		if((msg->flags[i] & SHMEM_ACTIVE) && (msg->flags[i] & SHMEM_OPENED)) {
		dlog(DEBUG_FILE,"[%d] [%s] -> [%s] opened",i,msg->from[i],msg->to[i]);
		if(msg->flags[i] & SHMEM_STATIOND) {
			stationd_handle(i);
			} else {
			shmem_ins_delete(i);
			}
		}
	}
	
	/* scan the to_delete queue for potential deletables */

	shmem_do_delete();
	return;
}

void shmem_ins_delete(int slot) {
	/* insert the delete onto the queue */
	struct parse *mt;
	struct msg *msg = filed->msg;
	struct todel *d;
	struct todel *e;
	int i;
	char *c;

	if(msg->flags[slot] & SHMEM_INS_DEL) return;
	msg->flags[slot] += SHMEM_INS_DEL;

	mt = get_media_type(msg->from[slot]);
	if(!mt) {
		 i = 0; 
	} else {
		if(!(c = get_token_s(mt,4))) i = 0;
		else
		i = atoi(c);
	}

	if(!(d = (struct todel *) malloc(sizeof(struct todel)))) {
		elog("shmem_ins_delete() error performing malloc");
		return;
	}

	d->slot = slot;
	d->delete_at = time(NULL) + i;
	
	if(filed->todel) {
		e = filed->todel;
		while(e->next) e = e->next;
		e->next = d;
		d->prev = e;
	} else {
		filed->todel = d;
		d->prev = NULL;
	}

	d->next = NULL;
	return;
}

void shmem_do_delete(void) {
	struct todel *d = filed->todel;
	struct todel *e;
	struct msg *msg = filed->msg;
	time_t now;


	/* perform deletion and remove */

	now = time(NULL);

	while(d) {
		e = NULL;
		if(d->delete_at <= now) {
			dlog(DEBUG_FILE,"performing removal of link in slot %d",d->slot);
			session_update(msg->id[d->slot]);
			shmem_file_remove(d->slot);
			if(d->prev) d->prev->next = d->next;
			if(d->next) d->next->prev = d->prev;
			if(d == filed->todel) 
				filed->todel = d->next;
			e = d;
		}
		d = d->next;
		if(e) free(e); /* raver's dream */
	}
	return;
}
